package org.gnurou.audioscrobbler;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.math.BigInteger;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * This class implements the Audioscrobbler 1.1 protocol.
 */
public class Audioscrobbler {
	public static boolean debug = false;
	private static final String lastFMURL = "http://post.audioscrobbler.com/?hs=true&p=1.1&c=tst&v=1.0";
	
	/// URL used for handshaking
	private String handshakeUrl;
	// User name used to authenticate
	private String username;	
	/// MD5 hash of the user password
	private String passwordHash;
	/// Response to the MD5 challenge of the server
	private String challengeResponse;
	/// URL that the server tells us to send our tracks to
	private String tracksURL;

	/// Timestamp of the latest request we did to the server
	private long lastRequest = 0;
	/// Number of seconds to wait from lastRequest before we can send
	/// another request to the server.
	private int interval = 0;
	
	/**
	 * Default constructor, submit songs to last.fm
	 */
	public Audioscrobbler(String username, String password) throws NoSuchAlgorithmException
	{
		this(lastFMURL, username, password);
	}

	/**
	 * Constructor allowing to give the URL of the audioscrobbler server.
	 */
	public Audioscrobbler(String url, String username, String password) throws NoSuchAlgorithmException
	{
		this.handshakeUrl = url;
		this.username = username;
		this.passwordHash = md5hash(password);
	}
	
	/**
	 * Authenticate to the audioscrobbler. This phase must be performed successfully before 
	 * doing any other action.
	 * @return true if authentication succeeded, false otherwise.
	 */
	public void handshake() throws BadUserError, ProtocolError, IOException,
		MalformedURLException, NoSuchAlgorithmException
	{
		URL url = new URL(handshakeUrl + "&u=" + URLEncoder.encode(username, "UTF-8"));
		waitAsServerRequested();
		InputStream is = url.openStream();
		BufferedReader br = new BufferedReader(new InputStreamReader(is));
		boolean success = false;
		// Read the response from the server
		String response = br.readLine();
		printDebug("Handshake response: " + response);
		// If our request succeeded, we have to read the md5 challenge as well as the 
		// URL to submit our tracks to.
		if (response.equals("UPTODATE") || response.startsWith("UPDATE"))
		{
			success = true;
			// Get the md5 challenge and compute our reply
			String md5challenge = br.readLine();
			printDebug("MD5 challenge: " + md5challenge);
			this.challengeResponse = md5hash(passwordHash + md5challenge);
			// Get the URL where we can submit our tracks to
			this.tracksURL = br.readLine();
		}
		// Read the interval
		String intervalString = br.readLine();
		printDebug("Interval string: " + intervalString);
		this.interval = getInterval(intervalString);
		this.lastRequest = getCurrentTime();
		
		// There were a failure... Let's throw the right exception
		if (!success)
		{
			if (response.equals("BADUSER")) throw new BadUserError();
			else throw new ProtocolError(getProtocolError(response));
		}		
	}

	/**
	 * Returns the number of seconds to wait before we can send a request to the
	 * server. It is safe to send a request when this method returns zero.
	 */
	public long getSecondsToWait()
	{
		long currentTime = getCurrentTime();
		long secsToWait = lastRequest + interval - currentTime;
		if (secsToWait < 0) return 0;
		return secsToWait;
	}
	
	/**
	 * Ensures that the delay requested by the server before issuing 
	 * another request is honoured.
	 *
	 */
	public void waitAsServerRequested()
	{
		boolean completed = false;
		long secsToWait = getSecondsToWait();
		do
		{
			try
			{
				if (secsToWait > 0) Thread.sleep(secsToWait * 1000);
				completed = true;
			} catch (InterruptedException e)
			{
				secsToWait = getSecondsToWait();
			}
		} while (!completed);
	}
	
	/**
	 * Returns the URL we can submit our played tracks to. The handshake
	 * must be successfully performed before this method returns something
	 * valuable.
	 */
	public String getTracksSubmitURL()
	{
		return tracksURL;
	}

	/**
	 * Submit a single track.
	 */
	public void submitTrack(AudioscrobblerTrack t) throws NoHandshakeError, MalformedURLException, 
		IOException, BadUserError, ProtocolError
	{
		if (tracksURL == null) throw new NoHandshakeError();
		
		// Open a connection to the submit URL
		URL url = new URL(tracksURL);
		URLConnection connection = url.openConnection();
		connection.setDoOutput(true);

		// Forge the URL data
		String data = "u=" + URLEncoder.encode(username, "UTF-8") +
			"&s=" + URLEncoder.encode(challengeResponse, "UTF-8") +
			t.getAudioscrobblerSubmissionString();
		
		waitAsServerRequested();
		// Send our request
		OutputStreamWriter wr = new OutputStreamWriter(connection.getOutputStream());
		printDebug("Sending track: " + data);
		wr.write(data);
		wr.flush();

		// Look at the server response
		BufferedReader br = new BufferedReader(new InputStreamReader(connection.getInputStream()));

		String response = br.readLine();
		printDebug("Server response: " + response);
		
		boolean success = false;
		if (response.equals("OK"))
			success = true;
		// Update the interval
		String intervalString = br.readLine();
		printDebug("Interval string: " + intervalString);
		interval = getInterval(intervalString);
		
		// In case of failure, throw the right exception
		if (!success)
		{
			if (response.equals("BADUSER")) throw new BadUserError();
			else throw new ProtocolError(getProtocolError(response));			
		}
	}
	
	/**
	 * Submit an array of tracks. This method tries to minimize the number
	 * of requests and to honor the wait time requires by the server between
	 * two requests.
	 */
	public void submitTracks(AudioscrobblerTrack[] tracks) throws NoHandshakeError, MalformedURLException, 
		IOException, BadUserError, ProtocolError
	{
		if (tracksURL == null) throw new NoHandshakeError();

		int processed = 0;
		while (processed < tracks.length)
		{
			// Open a connection to the submit URL
			URL url = new URL(tracksURL);
			URLConnection connection = url.openConnection();
			connection.setDoOutput(true);

			// Forge the URL data
			String data = "u=" + URLEncoder.encode(username, "UTF-8") +
				"&s=" + URLEncoder.encode(challengeResponse, "UTF-8");

			int i;
			for (i = processed; i < processed + 10 && i < tracks.length;
				i++)
				data += tracks[i].getAudioscrobblerSubmissionString(i - processed);
			
			waitAsServerRequested();
			// Send our request
			OutputStreamWriter wr = new OutputStreamWriter(connection.getOutputStream());
			printDebug("Sending tracks: " + data);
			wr.write(data);
			wr.flush();

			// Look at the server response
			BufferedReader br = new BufferedReader(new InputStreamReader(connection.getInputStream()));

			String response = br.readLine();
			printDebug("Server response: " + response);			
			
			boolean success = false;
			if (response.equals("OK"))
				success = true;
			// Update the interval
			String intervalString = br.readLine();
			printDebug("Interval string: " + intervalString);
			interval = getInterval(intervalString);
			
			// In case of failure, throw the right exception
			if (!success)
			{
				if (response.equals("BADUSER")) throw new BadUserError();
				else throw new ProtocolError(getProtocolError(response));			
			}
			processed = i;
		}
	}
	
	/**
	 * Given the interval string returned by the server, returns the interval at which
	 * we can submit requests in seconds.
	 */
	private int getInterval(String intervalString)
	{
		return Integer.parseInt(intervalString.substring("INTERVAL ".length()));
	}
	
	/**
	 * Given a "FAILED <reason>" String, returns the reason.
	 */
	private String getProtocolError(String errorString)
	{
		return errorString.substring("FAILED ".length());
	}
	
	/**
	 * Returns a correct timestamp of the current time.
	 * @return
	 */
	private static long getCurrentTime()
	{
		return System.currentTimeMillis() / 1000;
	}
	
	/**
	 * Returns a string representation of the md5 hash performed on the given
	 * string. All data must be in hexadecimal.
	 * 
	 * @throws NoSuchAlgorithmException if the JRE doesn't support MD5.
	 */
	private static String md5hash(String s) throws NoSuchAlgorithmException
	{
		MessageDigest md = MessageDigest.getInstance( "MD5" );
		md.update(s.getBytes());
		BigInteger hash = new BigInteger(1, md.digest());
		return hash.toString(16);
	}
	
	private static void printDebug(String s)
	{
		if (debug) System.err.println(s);
	}
}
