package org.gnurou.audioscrobbler;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.TimeZone;
import java.util.Vector;

import com.thoughtworks.xstream.XStream;

/**
 * Represents a musical collection, i.e. a set of tracks with
 * relevant information for each.
 */
public class Collection {
	protected HashMap<Track, TrackInfo> tracks;
	
	public Collection()
	{
		tracks = new HashMap<Track, TrackInfo>();
	}
	
	/**
	 * Get all the tracks to submit to the Audioscrobbler server, with play time information,
	 * sorted by play time.
	 * @param localCollection The local collection to compare our play counts with
	 * @param ignoreClock If true, the tracks play times are calculated from the
	 * lastPlayTime parameter. Otherwise, the play times returned by the Karma are used.
	 * @param lastPlayTime The latest play time of the tracks. Only makes sense if
	 * ignoreClock is true.
	 */
	public AudioscrobblerTrack[] getTracksToAdd(Collection localCollection, boolean ignoreClock, Date lastPlayTime)
	{
		Vector<AudioscrobblerTrack> updates = new Vector<AudioscrobblerTrack>();
		
		long minLastPlayTime;
		
		// The minimal play time initial value depends on whether
		// we are relying on the Karma clock or not.
		if (!ignoreClock) minLastPlayTime = Long.MAX_VALUE;
		else minLastPlayTime = lastPlayTime.getTime();
		
		Map<Track,Long> multiPlays = new HashMap<Track,Long>();
		
		for (Track track : tracks.keySet())
		{
			// Don't bother with tracks that last less than 30 seconds,
			// as the protocol requests.
			if (track.getLength() < 30) continue;
			
			// Also forget about tracks which have no title or artist, since
			// the server won't accept them.
			if (track.getTitle().equals("") || track.getArtist().equals("")) continue;
			
			TrackInfo info = tracks.get(track);
			TrackInfo localInfo = localCollection.tracks.get(track);

			// How many times has this track been played?
			long timesPlayed;
			if (localInfo == null) timesPlayed = info.getPlayCount();
			else timesPlayed = info.getPlayCount() - localInfo.getPlayCount();
			// If zero, then don't worry...
			if (timesPlayed == 0) continue;

			// If we don't ignore the clock, then set the play time to
			// what the Karma says...
			if (!ignoreClock)
			{
				// Fetch the last time this track was played...
				long date = info.getLastPlayTime();
				// ... and check if it's the oldest time we have.
				if (date < minLastPlayTime) minLastPlayTime = date;
				// Add the track as having been played at the time we know
				updates.add(new AudioscrobblerTrack(track, date - TimeZone.getDefault().getOffset(date)));
				// Schedule other appearances of this track for earlier scheduling.
				if (timesPlayed > 1)
					multiPlays.put(track, timesPlayed - 1);
			}
			// Otherwise let the multiplay do its job!
			else multiPlays.put(track, timesPlayed);
		}
		// Now set the earlier schedulings
		while (!multiPlays.isEmpty())
		{
			// Try to insert tracks in free holes
			for (Iterator it = multiPlays.keySet().iterator(); it.hasNext();)
			{
				Track track = (Track) it.next();
				long numPlays = multiPlays.get(track);
				if (numPlays == 0) { it.remove(); continue; }
				minLastPlayTime -= track.getLength() * 1000;
				updates.add(0, new AudioscrobblerTrack(track, minLastPlayTime - TimeZone.getDefault().getOffset(minLastPlayTime)));
				multiPlays.put(track, --numPlays);
			}
		}
		AudioscrobblerTrack[] tracks = updates.toArray(new AudioscrobblerTrack[updates.size()]); 
		Arrays.sort(tracks);
		// Suppress entries that last less than half of their playtime or 30 seconds
		updates = new Vector<AudioscrobblerTrack>();
		for (AudioscrobblerTrack t : tracks) updates.add(t);
		for (int i = 0; i < updates.size() - 1; i++)
			if (updates.get(i + 1).getDate() - updates.get(i).getDate() < (updates.get(i).getLength() / 2))
				updates.remove(i);
		// Make the final array and return it!
		tracks = updates.toArray(new AudioscrobblerTrack[updates.size()]); 
		return tracks;
	}
	
	/**
	 * Insert the replayed track into the given set. It will try to place it into
	 * one of the gaps of the set, or will put it at the beginning if no gap is found.
	 * 
	 * The set must be sorted by play time.
	 */
	private void scheduleReplayedTrack(Track track, Vector<AudioscrobblerTrack> tracks)
	{
		
	}
	
	public void save(String fileName) throws FileNotFoundException, IOException
	{
		XStream xstream = createXStream();
		ObjectOutputStream out = xstream.createObjectOutputStream(new OutputStreamWriter(new FileOutputStream(fileName)));
		out.writeObject(tracks);
		out.close();
	}
	
	public void load(String fileName) throws IOException, FileNotFoundException, 
		ClassNotFoundException
	{
		XStream xstream = createXStream();
		ObjectInputStream in = xstream.createObjectInputStream(new InputStreamReader(new FileInputStream(fileName)));
		tracks = (HashMap<Track, TrackInfo>) in.readObject();
		in.close();
	}
	
	public void addTrack(Track t, TrackInfo i)
	{
		tracks.put(t, i);
	}
	
	private XStream createXStream()
	{
		XStream xstream = new XStream();
		//xstream.setMode(XStream.ID_REFERENCES);
		xstream.alias("track", Track.class);
		xstream.alias("info", TrackInfo.class);
		return xstream;
	}
	
	public Track[] getTracks() 
	{
		return tracks.keySet().toArray(new Track[tracks.size()]);
	}
}
