package org.gnurou.karmascrobbler;
/**
 *  Copyright (C) 2006 Alexandre Courbot.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.TimeZone;

import org.biffle.cliff.pearl.PearlException;
import org.gnurou.audioscrobbler.Audioscrobbler;
import org.gnurou.audioscrobbler.AudioscrobblerTrack;
import org.gnurou.audioscrobbler.Collection;
import org.gnurou.audioscrobbler.RioCollection;

public class Main {

	private final static String version = "0.1.4";
	
	private static String rioIP;
	private static int rioPort = RioCollection.defaultRioPort;
	private static String rioPass;
	private static String localCollectionFile = "collection.xml";

	private static String lastFMUser;
	private static String lastFMPass;
	
	private static Date latestDate = null;
	private static boolean ignoreClock = false;
	
	private static RioCollection rioCollection = new RioCollection();
	private static Collection localCollection = new Collection();

	/**
	 * Displays the program usage and exit.
	 *
	 */
	private static void usage()
	{
		System.out.println("karmascrobbler version " + version);
		System.out.println("Usage: java -jar karmascrobbler.jar -ip karma_ip -lfmuser last_fm_user -lfmpass last_fm_pass [ -riopass karma_password -rioport karma_port -collection collection_file -forcedate \"YYYY-MM-DD hh:mm:ss\" -debug]");
		System.exit(0);
	}
	
	/**
	 * Parse command-line options.
	 */
	private static void parseOptions(String[] args)
	{
		for (int idx = 0; idx < args.length; idx++)
		{
			if (args[idx].equals("-debug"))
			{
				Audioscrobbler.debug = true;
			}
			else if (args[idx].equals("-ip") && idx < args.length - 1)
			{
				rioIP = args[++idx];
			}
			else if (args[idx].equals("-port") && idx < args.length - 1)
			{
				rioPort = Integer.parseInt(args[++idx]);
			}
			else if (args[idx].equals("-riopass") && idx < args.length - 1)
			{
				rioPass = args[++idx];
			}
			else if (args[idx].equals("-collection") && idx < args.length - 1)
			{
				localCollectionFile = args[++idx];
			}
			else if (args[idx].equals("-lfmuser") && idx < args.length - 1)
			{
				lastFMUser = args[++idx];
			}
			else if (args[idx].equals("-lfmpass") && idx < args.length - 1)
			{
				lastFMPass = args[++idx];
			}
			else if (args[idx].equals("-forcedate") && idx < args.length - 1)
			{
				SimpleDateFormat sfd = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
				try {
					Date d = sfd.parse(args[++idx]);
					latestDate = new Date(d.getTime() - TimeZone.getDefault().getOffset(d.getTime()));
				}catch (ParseException e)
				{
					System.out.println("Wrong date format!");
					usage();
				}
			}
			else if (args[idx].equals("-ignoreclock"))
			{
				ignoreClock = true;
			}
			else usage();
		}
	}

	/**
	 * Connect to the Karma, using the right method depending on whether
	 * a password has been given or not.
	 */
	private static void connect() throws PearlException, IOException
	{
		if (rioPass != null)
			rioCollection.connect(rioIP, rioPort, rioPass);
		else rioCollection.connect(rioIP, rioPort);
	}
	
	public static void main(String[] args) {
		parseOptions(args);

		// These 3 variables are strictly mandatory
		if (rioIP == null) usage();
		if (lastFMUser == null) usage();
		if (lastFMPass == null) usage();
		// Requesting to ignore the Karma clock without giving a
		// base date does not make sense.
		if (ignoreClock && latestDate == null)
		{
			System.out.println("Ignoring the Karma clock requires you to specify a base date with the -forcedate option!");
			usage();
		}
		
		// Now try to open the collection
		try {
			localCollection.load(localCollectionFile);
		}
		catch (FileNotFoundException e)
		{
			// Cannot find the collection? Let's create it then.
			System.out.println("Cannot find the collection file - I'm going to create" +
					" it from the Karma collection...");
			try {
				connect();
				rioCollection.buildCollection();
				rioCollection.closeConnection();
				rioCollection.save(localCollectionFile);
				System.out.println("Collection file saved - next time I'm run on it," +
						" I'll sync your last.fm entry!");
				System.exit(0);
			}catch (Exception exc)
			{
				exc.printStackTrace();
				System.exit(1);
			}			
		}
		catch (Exception e)
		{
			e.printStackTrace();
			System.exit(1);
		}
		
		// We have our local collection loaded, we can now fetch the RIO collection!
		try {
			System.out.println("Connecting to the Karma...");
			connect();
			rioCollection.buildCollection();
			rioCollection.closeConnection();
			// All the tracks that shall be added
			AudioscrobblerTrack[] tracks = rioCollection.getTracksToAdd(localCollection, ignoreClock, latestDate);
			if (tracks.length == 0)
			{
				System.out.println("Sorry, no track to add to your profile!");
				System.exit(0);
			}
			
			// Check whether the dates should be forced, and adapt them
			if (latestDate != null)
			{
				long timediff = latestDate.getTime() - tracks[tracks.length - 1].getDate();
				for (AudioscrobblerTrack track : tracks)
				{
					track.adjustDate(timediff);
				}
			}
			
			System.out.println("I'm going to add the following tracks to your profile:");
			for (AudioscrobblerTrack track : tracks)
			{
				System.out.println("* " + track.getTitle() + " by " + track.getArtist() + ", played on " + track.formatDateToLocalTime());
			}

			// Expect user confirmation
			InputStreamReader isr = new InputStreamReader(System.in);
			BufferedReader reader = new BufferedReader(isr);
			do {
				System.out.print("Shall I proceed? [Y/N] ");
				System.out.flush();
				String line = reader.readLine();
				if (line.equals("Y") || line.equals("y")) break;
				if (line.equals("N") || line.equals("n")) System.exit(0);
			} while (true);
			
			System.out.println("Submitting the tracks to Last.fm...");
			Audioscrobbler scrobbler = new Audioscrobbler(lastFMUser, lastFMPass);
			scrobbler.handshake();
			scrobbler.submitTracks(tracks);
			
			System.out.println("Updating your local collection...");
			rioCollection.save(localCollectionFile);
			System.out.println("Successfully completed!");
		}
		catch (Exception e)
		{
			e.printStackTrace();
			System.exit(1);
		}
	}
}
